/* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * C implementation for Std module.
 */

#include <gont/runtime.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

extern gont_ptroff Std$$Match_failure;
extern gont_ptroff Std$$Null_access;
extern gont_ptroff Std$$End_of_file;

struct gont_string *gont_cstring(const char *s)
{
	return gont_make_string(strlen(s), s);
}

void Std__print_string(struct gont_string *s)
{
	fwrite(s->data, s->len, 1, stdout);
}

void Std__prerr_string(struct gont_string *s)
{
	fwrite(s->data, s->len, 1, stderr);
}

/* TODO: fix this to get longer lines. */
struct gont_string *Std__read_line_f(FILE *f)
{
	char buf[4096];
	char *r;

	r = fgets(buf, sizeof(buf), f);

	if (r == NULL)
		gont_raise(gont_make_union(Std$$End_of_file, NULL));
	
	return gont_cstring(buf);
}

struct gont_string *Std__read_line(void)
{
	return Std__read_line_f(stdin);
}

struct gont_string *Std__itoa(int i)
{
	char buf[100];

	sprintf(buf, "%d", i);
	
	return gont_cstring(buf);
}

struct gont_string *Std__ftoa(double *f)
{
	char buf[100];

	sprintf(buf, "%g", *f);
	
	return gont_cstring(buf);
}

void gont_match_failure(void)
{
	gont_raise(gont_make_union(Std$$Match_failure, NULL));
}

void gont_null_access(void)
{
	gont_raise(gont_make_union(Std$$Null_access, NULL));
}

gont_ptroff Std__atoi(struct gont_string *s)
{
	char *p;
	gont_ptroff r;

	if (s->len == 0)
		goto fail;
		
	r = strtol(s->data, &p, 0);

	while (isspace(*p))
		p++;
		
	if (*p != '\0')
		goto fail;
	
	return r;

fail:
	gont_invalid_argument("Std::atoi");
	return -1;	/* never reached */
}

double *Std__atof(struct gont_string *s)
{
	char *p;
	double r;

	if (s->len == 0)
		goto fail;
		
	r = strtod(s->data, &p);

	while (isspace(*p))
		p++;
		
	if (*p != '\0')
		goto fail;
	
	return gont_make_float(r);

fail:
	gont_invalid_argument("Std::atof");
	return NULL;	/* never reached */
}

void Std__exit(gont_ptroff x)
{
	gont_run_exit_routines();
	exit(x);
}

void Std__at_exit(struct gont_lambda_exp *fnc)
{
	gont_register_exit_routine((void (*)(void*))fnc->fnc, fnc->closure);
}

gont_ptroff gont_streq(struct gont_string *s1, struct gont_string *s2)
{
	return s1->len == s2->len &&
	       memcmp(s1->data, s2->data, s1->len) == 0;
}

gont_ptroff gont_strcmp(struct gont_string *s1, struct gont_string *s2)
{
	int r;
	
	if (s1->len < s2->len) {
		r = memcmp(s1->data, s2->data, s1->len);
		return r == 0 ? -1 : r;
	} else if (s1->len > s2->len) {
		r = memcmp(s1->data, s2->data, s2->len);
		return r == 0 ? 1 : r;
	} else
		return memcmp(s1->data, s2->data, s1->len);
}

struct gont_string *gont_strjoin(struct gont_string *s1, 
			         struct gont_string *s2)
{
	struct gont_string *r = gont_make_string(s1->len + s2->len, NULL);

	memcpy(r->data, s1->data, s1->len);
	memcpy(r->data + s1->len, s2->data, s2->len);

	return r;
}

gont_ptroff gont_id(gont_ptroff x)
{
	return x;
}

gont_ptroff gont_hash_string(struct gont_string *s)
{
	gont_ptroff pos = 0, hash = 0;

	while (pos < s->len) {
		hash += s->data[pos++];
		hash *= 37;
	}

	return hash & 0x7fffffff;
}
